#if defined(HST_NEWTEX4)
# define BIN_DIM 64
# define BIN_STEPS 2
# define BIN_STEP 8
# define PROJ_DIM 4
# if defined(HST_FLOAT_LOOPS)
#  define PROJ_STEP 4.f
# else
#  define PROJ_STEP 4
# endif
#else
# define BIN_DIM 16
# define BIN_STEPS 4
# define BIN_STEP 4
# define PROJ_DIM 16
# if defined(HST_FLOAT_LOOPS)
#  define PROJ_STEP 16.f
# else
#  define PROJ_STEP 16
# endif
#endif

#if SLICE_BLOCK == 4
# define shfl_sum(val, k) ({ \
    val.x += __shfl_xor(val.x, k, PROJ_STEP); \
    val.y += __shfl_xor(val.y, k, PROJ_STEP); \
    val.z += __shfl_xor(val.z, k, PROJ_STEP); \
    val.w += __shfl_xor(val.w, k, PROJ_STEP); \
})
#elif SLICE_BLOCK == 2
# define shfl_sum(val, k) ({ \
    val.x += __shfl_xor(val.x, k, PROJ_STEP); \
    val.y += __shfl_xor(val.y, k, PROJ_STEP); \
})
#else
# define shfl_sum(val, k) ({ \
    val += __shfl_xor(val, k, PROJ_STEP); \
})
#endif



#ifdef HYBRID_KEPLER
__device__
#else
__global__ 
#endif
static 
#if defined(HST_SET_BOUNDS) && !defined(HYBRID_KEPLER)
__launch_bounds__(256, HST_NEWTEX_MIN_BLOCKS)
#endif /* HST_SET_BOUNDS */
void 
#ifdef HYBRID_KEPLER
hst_cuda_linear_companion
#else 
hst_kepler_kernel
#endif
#if defined(HST_FLOAT_LOOPS)
(cudaTextureObject_t texptr, const float * __restrict__ g_all, float num_proj, int num_bins, vfloat *d_SLICE, float apos_off_x, float apos_off_y,
#else
(cudaTextureObject_t texptr, const float * __restrict__ g_all, int num_proj, int num_bins, vfloat *d_SLICE, float apos_off_x, float apos_off_y,
#endif
#ifdef HYBRID_KEPLER
const int bidx, const int bidy) {
#else
int batch) {
#endif
    float h;

    vfloat res[BIN_STEPS][BIN_STEPS] = {0};
#if  (defined(HST_SHFL_SUM) && (SLICE_BLOCK == 1))
    __shared__ vfloat buf[BIN_DIM][PROJ_DIM + 4]; 
    __shared__ vfloat fin[16][16];
#else
    __shared__ vfloat buf[BIN_DIM][PROJ_DIM]; 
    __shared__ vfloat fin[16][16];
#endif

    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

#ifndef HYBRID_KEPLER
    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;
#endif

    const int proj_minor = tidx / 4;
    const int xy_minor = tidx % 4;
    const int y_minor = xy_minor / 2;
    const int x_minor = xy_minor % 2;

    const int xy_major = tidy % 4;
    const int y_major = xy_major / 2;
    const int x_major = xy_major % 2;

#if defined(HST_NEWTEX4)
    const int block = tidy / 4;
    const int blocky = block/2;
    const int blockx = block%2;

    const int ridy = 4 * tidy + tidx / 4;
    const int ridx = tidx % 4;

    const int sidy = blocky * 4 + y_major * 2 + y_minor;
    const int sidx = blockx * 4 + x_major * 2 + x_minor;

    const int proj = proj_minor;
#else
    const int ridy = tidy;
    const int ridx = tidx;

    const int sidy = y_major * 2 + y_minor;
    const int sidx = x_major * 2 + x_minor;

    const int proj_major = tidy / 4;
    const int proj =  4 * proj_major + proj_minor;
#endif

    const float x = bidx + sidx + apos_off_x;
    const float y = bidy + sidy + apos_off_y;


#if defined(HST_FLOAT_LOOPS)
    for (float projf = proj + 0.5f; projf < num_proj; projf += PROJ_STEP) {
//        float fidx = projf + exp2(23.f); const int idx = (*(int*)(&fidx)) - 0x4B000000;
        const int proje = (int)projf;
#else
    for (int proje = proj; proje < num_proj; proje += PROJ_STEP) {
	const float projf = proje + 0.5f;
#endif
        const float4 all = c_all[proje];		// (int)projf (or fancy round) to avoid register spills
        h = all.z + x * all.x - y * all.y;

#pragma unroll
        for (int i = 0; i < BIN_STEPS; i++) {
#pragma unroll
	    for (int j = 0; j < BIN_STEPS; j++) {
	        float subh = h + BIN_STEP * j * all.x - BIN_STEP * i * all.y;
	        res[i][j] += hst_tex(texptr, subh, projf);
	    }
        }
    }

#pragma unroll
    for (int i = 0; i < BIN_STEPS; i++) {
#pragma unroll
	for (int j = 0; j < BIN_STEPS; j++) {
            buf[BIN_STEP * sidy + sidx][proj] = res[i][j];

            __syncthreads();

#ifdef HST_SHFL_SUM
            vfloat val = buf[ridy][ridx];
# pragma unroll
            for (int k=(PROJ_DIM/2); k>=1; k/=2) 
        	shfl_sum(val, k);
#else // HST_SHFL_SUM
# pragma unroll
            for (int k=(PROJ_DIM/2); k>=1; k/=2) {
                buf[ridy][ridx] += buf[ridy][ridx+k];
            	    // We need it instead of volatile. Could be faster if buf declared volatile
                __syncthreads();
            }
            vfloat val = buf[ridy][0];
#endif //  HST_SHFL_SUM

            const int rx = BIN_STEP * j + ridy%BIN_STEP;
            const int ry = BIN_STEP * i + ridy/BIN_STEP;

            if (!ridx) {
                fin[ry][rx] = val;
            }

            __syncthreads();
        }
    }


    const int idx = bidx + tidx;
    const int idy = bidy + tidy;

#ifdef HYBRID_KEPLER
    d_SLICE[PPT * BLOCK_SIZE_X * gridDim.x * idy + idx] = fin[tidy][tidx];
#else
    d_SLICE[BLOCK_SIZE_X * gridDim.x * idy + idx] = fin[tidy][tidx];
#endif
}
