/*
 * The PyHST program is Copyright (C) 2002-2011 of the
 * European Synchrotron Radiation Facility (ESRF) and
 * Karlsruhe Institute of Technology (KIT).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * hst is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HST_C_SIN
__constant__ float  c_sin[ MAXNPROJECTIONS  ] ;	//!< (cosinuses, sinuses, axis position, min bin
#endif
#ifdef HST_C_TRIG
__constant__ float2  c_trig[ MAXNPROJECTIONS  ] ;	//!< (cosinuses, sinuses, axis position, min bin
__constant__ float2  c_ofst[ MAXNPROJECTIONS  ] ;	//!< (cosinuses, sinuses, axis position, min bin
#endif
__constant__ float4  c_all[ MAXNPROJECTIONS  ] ;	//!< (cosinuses, sinuses, axis position, min bin
texture<float, 2, cudaReadModeElementType> tex_projes;


#include "hst_cuda_fai.h"


#if SLICE_BLOCK == 4
# define vfloat float4
#elif SLICE_BLOCK == 2
# define vfloat float2
#else 
# define vfloat float
#endif

#ifndef HST_CREATE_TEXTURE
texture<vfloat, 2, cudaReadModeElementType> array_projes;
#endif /* !HST_CREATE_TEXTURE */

#if defined(HST_CUDA_ARRAY)
# ifdef HST_CREATE_TEXTURE
#  define hst_tex(t, x, y) tex2D<vfloat>(t, x, y)
# else 
#  define hst_tex(t, x, y) tex2D(array_projes, x, y)
# endif
#else
# define hst_tex(t, x, y) tex2D(tex_projes, x, y)
#endif


#define comp0 x
#define comp1 y
#define comp2 z
#define comp3 w
#define comp(i) comp##i

#define mkf4(xy, zw) make_float4(xy.x, xy.y, zw.x, zw.y)

#define HST_MACRO_JOIN(a, b) a ## b

#include <math_constants.h>
#define M_PI_F CUDART_PI_F



__global__ static void hst_cuda_kernel(cudaTextureObject_t texptr, int num_proj, int num_bins, vfloat *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    const int idx = blockIdx.x * BLOCK_SIZE_X + threadIdx.x;
    const int idy = blockIdx.y * BLOCK_SIZE_Y + threadIdx.y + batch;

    float h;
    vfloat res={0};

    const float bx = idx + apos_off_x;
    const float by = idy + apos_off_y;

    for (int proje=0; proje<num_proj; proje++) {
        const float4 all = c_all[proje];
        h = all.z + bx * all.x - by * all.y;

        res=res+hst_tex(texptr, h, proje + 0.5f) ;
    }

    d_SLICE[ BLOCK_SIZE_X*gridDim.x*idy + idx ] = res;
}


#include "hst_cuda_bp_kepler.h"


#define PROJ_BLOCK HST_LINEAR_PROJ_BLOCK
#define PPT HST_LINEAR_PPT
#define BS HST_LINEAR_BLOCK
#ifdef HST_LINEAR_MPLINEAR
# include "hst_cuda_bp_mplinear.h"
#else
# include "hst_cuda_bp_simdlinear.h"
#endif
#undef BS
#undef PPT
#undef PROJ_BLOCK


#define OVERSAMPLE 4
#define PROJ_BLOCK HST_OVERS_PROJ_BLOCK
#define PPT HST_OVERS_PPT
#define BS HST_LINEAR_BLOCK
#ifdef HST_LINEAR_MPLINEAR
# include "hst_cuda_bp_mplinear.h"
#else
# include "hst_cuda_bp_simdlinear.h"
#endif
#undef BS
#undef PPT
#undef PROJ_BLOCK
#undef OVERSAMPLE


#define OVERSAMPLE 1
#define PROJ_BLOCK HST_NN_PROJ_BLOCK
#define PPT HST_NN_PPT
#define BS HST_LINEAR_BLOCK
#ifdef HST_LINEAR_MPLINEAR
# include "hst_cuda_bp_mplinear.h"
#else
# include "hst_cuda_bp_simdlinear.h"
#endif
#undef BS
#undef PPT
#undef PROJ_BLOCK
#undef OVERSAMPLE



#ifdef HST_HYBRID
# define PPT HST_LINEAR_PPT
# ifdef HST_HYBRID_NEWTEX
#  define HYBRID_KEPLER
#  include "hst_cuda_bp_kepler.h"
#  undef HYBRID_KEPLER
# else

__device__ static void hst_cuda_linear_companion(cudaTextureObject_t texptr, const float * __restrict__ g_all, int num_proj, int num_bins, vfloat *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    vfloat res[PPT][PPT] = {0};

    const int sidx = 4 * (threadIdx.y % 4) + (threadIdx.x % 4);
    const int sidy = 4 * (threadIdx.y / 4) + (threadIdx.x / 4);

//    const int sidx = threadIdx.x;
//    const int sidy = threadIdx.y;

    const int idx = PPT * blockIdx.x * BLOCK_SIZE_X + sidx;
    const int idy = PPT * blockIdx.y * BLOCK_SIZE_Y + sidy + batch;

    float h;
    const float bx = idx + apos_off_x;
    const float by = idy + apos_off_y;

	// DS: float loops?
    for (int proje=0; proje<num_proj; proje++) {
        const float4 all = c_all[proje];
        h = all.z + bx * all.x - by * all.y;

#pragma unroll
        for (int i = 0; i < PPT; i++) {
#pragma unroll
            for (int j = 0; j < PPT; j++) {

                res[i][j] += hst_tex(texptr, h + j * BLOCK_SIZE_X * all.x - i * BLOCK_SIZE_Y * all.y, proje + 0.5f) ;
            }
        }
    }


#pragma unroll
    for (int i = 0; i < PPT; i++) {
#pragma unroll
        for (int j = 0; j < PPT; j++) {
            d_SLICE[PPT * BLOCK_SIZE_X * gridDim.x * (idy + i * BLOCK_SIZE_Y) + idx + j * BLOCK_SIZE_X] = res[i][j];
        }
    }
}

# endif


__device__ uint get_smid(void) {
     uint ret;
     asm("mov.u32 %0, %smid;" : "=r"(ret) );
     return ret;

}

__device__ uint smblocks[128] = {0};


# if defined(HST_SET_BOUNDS)
__launch_bounds__(256, HST_NEWTEX_MIN_BLOCKS)
# endif /* HST_SET_BOUNDS */
__global__ static void hst_cuda_linear_hybrid(cudaTextureObject_t texptr, const float * __restrict__ g_all, const vfloat * __restrict__ sino, int num_proj, int num_bins, int bin_pitch, vfloat *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
//    int block = (blockIdx.y * gridDim.x +  blockIdx.x)/16;


    __shared__ uint block;
    if ((threadIdx.x == 0)&&(threadIdx.y == 0)) {
	uint smid = get_smid();
	block = atomicAdd(&smblocks[smid], 1);
    }
    __syncthreads();

    if ((block&0x7)<4) {
        hst_cuda_alu_linear(texptr, g_all, sino, num_proj, num_bins, bin_pitch, d_SLICE, apos_off_x, apos_off_y, batch);
    } else {
#ifdef HST_HYBRID_NEWTEX
	for (int i = 0; i < PPT; i++) {
	    for (int j = 0; j < PPT; j++) {
		const int bidx = (PPT * blockIdx.x + j) * HST_LINEAR_BLOCK;
		const int bidy = batch + (PPT * blockIdx.y + i) * HST_LINEAR_BLOCK;

		hst_cuda_linear_companion(texptr, g_all, num_proj, num_bins, d_SLICE, apos_off_x, apos_off_y, bidx, bidy);
	    }
	}
#else
	hst_cuda_linear_companion(texptr, g_all, num_proj, num_bins, d_SLICE, apos_off_x, apos_off_y, batch);
#endif
    }

}
# undef PPT
#endif





